
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */


#ifndef __INDICATOR_OBJECT_ENUM_TYPES_H__
#define __INDICATOR_OBJECT_ENUM_TYPES_H__

#include <glib-object.h>

G_BEGIN_DECLS
/* enumerations from "indicator-object.h" */

GType indicator_scroll_direction_get_type (void) G_GNUC_CONST;
#define INDICATOR_OBJECT_TYPE_SCROLL_DIRECTION (indicator_scroll_direction_get_type())
G_END_DECLS

#endif /* !__INDICATOR_OBJECT_ENUM_TYPES_H__ */

/* Generated data ends here */

